package game;

import com.jogamp.opengl.GL2;

import graphics.GFX;
import calc.Calc;
import calc.V3D;
import mechanics.Entity;

public class SphereEntity extends Entity
	{
	/**The local motion coordinate system. 
	 * Normally automatically set upon collisions. 
	 * The rest is up to implementations.*/
	public V3D x,y,z;
	/**Workspace*/
	public V3D c0;
	/**The collision sphere's radius*/
	public double radius;
	/**If the object is withing this distance from an object, a collision will be detected.
	 * Normally 0.*/
	public double distTolerance;
	/**Flag set to 1 when the entity touches a solid object on this step. 
	 * Implementations are responsible for handling the value beyond this.*/
	public float gnd;
	/**Friction coefficient*/
	public double frict;
	/**Gravity vector*/
	public V3D gravity;
	/**A number representing the scale factor for any models related to this object.*/
	public double scale;
	/**Used in collisions with other sphereEntities and such. Must not be 0*/
	public double mass;
	/**Log (positive) normal forces*/
	public double normalLog;
	
	public SphereEntity(V3D pos)
		{
		super(pos);
		x = new V3D(1,0,0);
		y = new V3D(0,1,0);
		z = new V3D(0,0,1);
		c0 = new V3D();
		gnd=0;
		mass = 1;
		frict=1;
		normalLog=0;
		distTolerance=0;
		radius=16;
		scale=radius;
		gravity = new V3D(0,0.3,0);
		}
	
	@Override
	public void step()
		{
		super.step();
		speed.add(gravity,1);
		}
	@Override
	public void endStep()
		{
		}
	@Override
	public void render()
		{
		GL2 gl = GFX.gl;
		gl.glColor4fv(color,0);
		gl.glPushMatrix();
		gl.glTranslated(pos.x(), pos.y(), pos.z());
		gl.glScaled(scale, scale, scale);
		GFX.unitSphere2.render();
		gl.glPopMatrix();
		}
	/**
	 * Collision handling method. (Not collision detection.) 
	 * @param referencePoint The point of the collision.
	 * @param referenceSpeed The speed of the object we are colliding with.
	 * @param normal The normal vector for the collision.
	 * @param referenceRadius Typically 0. The new position will be referencePoint+normal*(referenceRadius+radius).
	 * */
	public void collide(V3D referencePoint, V3D referenceSpeed,V3D normal,double referenceRadius,double friction)	
		{
		referenceSpeed = speed.clone().sub(referenceSpeed);
		double d = -normal.dot(referenceSpeed);
		pos = referencePoint.clone().add(normal,referenceRadius+radius);
		if(d>0)
			normalLog+=d;
		speed.add(normal,d);
		double sp = speed.len();
		if((sp>0 || !z.equals(normal)) && (gnd<1 || normal.z()>z.z()))//If there is a choice, choose the z with highest z.
			{
			z = normal;
			if(sp==0)//This kinda sucks but we have to deal with it
				{
				if(x.isMultipleOf(z))//This practically never happens
					{
					x.setCross(y,z);
					y.setCross(z,x).normalize();
					x.setCross(y,z);
					}
				else//This should keep the direction for all practical appliances.
					{
					y.setCross(z,x);
					x.setCross(y,z).normalize();
					y.setCross(z,x);
					}
				}
			else//This is wonderful.
				{
				x = speed.clone().multiply(1d/sp);
				y.setCross(z,x);
				}
			}
		gnd=1;
		speed.accelerate(-d*friction*frict);
		}
	
	/**
	 * @param center The ball center
	 * @param ballRadius The ball radius
	 * @param referenceSpeed The ball speed
	 * @param ballFriction The friction of the surface. Typically 0.02-ish
	 * @param testWithSpeed Whether to include the speed when checking the distance
	 * */
	public void collisionBall(V3D center, double ballRadius, V3D referenceSpeed,double ballFriction, boolean testWithSpeed)
		{
		V3D daff = pos.clone().sub(center);
		double dist =daff.len();
		daff.multiply(1d/dist);
		//Note: workaround for hair.
		if(dist<=ballRadius+radius+getDistTolerance(daff) +(testWithSpeed?referenceSpeed.clone().sub(speed).dot(daff):0))
			{
			collide(center,referenceSpeed,daff,ballRadius,ballFriction);
			}
		}
	
	
	public void collisionSphereEntity(SphereEntity other)
		{
		V3D daff = pos.clone().sub(other.pos);
		double dist =daff.len();
		daff.multiply(1d/dist);
		double d = other.speed.clone().sub(speed).dot(daff);
		//Note: workaround for hair.
		if(dist<=other.radius+radius + d)
			{
			double weight;
			if(other.mass>mass)
				weight = 0.5*mass/other.mass;
			else
				weight = 1-0.5*other.mass/mass;
			V3D centerPos = pos.clone().multiply(weight).add(other.pos,1-weight);
			pos.set(centerPos).add(daff,(other.radius+radius)*(1-weight));
			other.pos.set(centerPos).add(daff,-(other.radius+radius)*weight);
			speed.add(daff,d*(1-weight));
			other.speed.add(daff,-d*weight);
			
			this.sphereEntityCollide(other);
			other.sphereEntityCollide(this);
			}
		}
	/**This indicates a reciprocal collision between this and other. 
	 * The actual collision is handled in collisionSphereEntity(which needs to be symmetric)
	 * This should handle eventual side effects.*/
	public void sphereEntityCollide(SphereEntity other)
		{
		// TODO Auto-generated method stub
		
		}

	protected double getDistTolerance(V3D normal)
		{
		return distTolerance;
		}

	/**
	 * @param p0 First point on triangle
	 * @param s0 Speed of first point on triangle
	 * @param p1 Second point on triangle
	 * @param s1 Speed of second point on triangle
	 * @param p2 Third point on triangle
	 * @param s2 Speed of third point on triangle
	 * @param normal The normal. Will be computed if null.
	 * */
	public void collisionTriangle(V3D p0, V3D s0, V3D p1, V3D s1, V3D p2, V3D s2, V3D normal, double triangleFriction)
		{
		// Compute vectors and stuff  
		V3D v0 =  p2.clone().sub(p0);
		double sqLen0 = v0.sqLen();
		V3D v1 =  p1.clone().sub(p0);
		double sqLen1 = v1.sqLen();
		
		//Project the point down on the plane
		if(normal==null)
			{c0.setCross(v0,v1).normalize();
			normal = c0;}
		V3D relativePos = pos.clone().sub(p0);
		V3D pointOnPlane = pos.clone().add(normal,-normal.dot(relativePos));
		V3D v2 = pointOnPlane.clone().sub(p0);
		
		// Compute dot products
		double dot00 = v0.dot(v0);
		double dot01 = v0.dot(v1);
		double dot02 = v0.dot(v2);
		double dot11 = v1.dot(v1);
		double dot12 = v1.dot(v2);
		// Compute barycentric coordinates
		double invDenom = 1d / (dot00 * dot11 - dot01 * dot01);
		double u = (dot11 * dot02 - dot01 * dot12) * invDenom;
		double v = (dot00 * dot12 - dot01 * dot02) * invDenom;
		
		V3D closestPoint; 
		V3D referenceSpeed;
		boolean testWithSpeed = true;
		// Check if point is in triangle
		if ((u >= 0) && (v >= 0) && (u + v <= 1))
			{
			referenceSpeed = s0.clone().multiply(1-u-v).add(s1.clone().multiply(v)).add(s2.clone().multiply(u));
			closestPoint = pointOnPlane;
			}
		else if(v<0)//It's on the 0-1 edge
			{
			testWithSpeed = false;
			double c = Calc.limit(dot02/(sqLen0), 0, 1);
			referenceSpeed = s0.clone().multiply(1-c).add(s1.clone().multiply(c));
			closestPoint = p0.add(v0,c);
			}
		else if (u<0)//It's on the 0-2 edge
			{
			testWithSpeed = false;
			double c = Calc.limit(dot12/(sqLen1), 0, 1);
			referenceSpeed = s0.clone().multiply(1-c).add(s2.clone().multiply(c));
			closestPoint = p0.add(v1,c);
			}
		else//This case is more complicated because we haven't computed enough data.
			{
			testWithSpeed = false;
			V3D vec = p2.sub(p1);
			double sqLen = vec.sqLen();
			double c = Calc.limit(vec.dot(pointOnPlane.clone().sub(p1))/(sqLen), 0, 1);
			referenceSpeed = s1.clone().multiply(1-c).add(s2.clone().multiply(c));
			closestPoint = p1.add(vec,c);
			}
		
		collisionBall(closestPoint, 0, referenceSpeed, triangleFriction,testWithSpeed);
		}
	}
