package graphics;

/**
THIS PROGRAM IS PROVIDED "AS IS" WITHOUT ANY WARRANTIES (OR CONDITIONS),
EXPRESS OR IMPLIED WITH RESPECT TO THE PROGRAM, INCLUDING THE IMPLIED WARRANTIES (OR CONDITIONS)
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE ENTIRE RISK ARISING OUT OF USE OR
PERFORMANCE OF THE PROGRAM AND DOCUMENTATION REMAINS WITH THE USER.
 */
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Insets;
import java.awt.LayoutManager2;
import java.util.HashMap;

/**
 *
 * A vertical layout manager similar to {@link FlowLayout}. Unlike FlowLayout, 
 * components will not wrap to form another column if there isn't enough space 
 * vertically. Like FlowLayout components do not expand to fill available space 
 * by default. However, The horizontal alignment may be set to <code>BOTH</code> 
 * in order to stretch the components horizontally.  VerticalLayout can optionally
 * anchor components to the top, bottom or center of the display area, or it can 
 * expand components vertically using weight constraints, e.g. passing a
 * {@link Double} as the constraint to {@link Container#add(Component, Object)}.
 * <br><br>
 * Implementation last revised on: May 19th 2021
 *
 * @author Oddwarg
 * @author Colin Mummery e-mail: colin_mummery@yahoo.com
 *         Homepage:www.kagi.com/equitysoft - Based on 'FlexLayout' in Java
 *         class libraries Vol 2 Chan/Lee Addison-Wesley 1998
 */

public class VerticalLayout implements LayoutManager2
{

	/**
	 * The horizontal alignment constant that designates centering. Also used to
	 * designate center anchoring.
	 */
	public final static int CENTER = 0;
	/**
	 * The horizontal alignment constant that designates right justification.
	 */
	public final static int RIGHT = 1;
	/**
	 * The horizontal alignment constant that designates left justification.
	 */
	public final static int LEFT = 2;
	/**
	 * The horizontal alignment constant that designates stretching the component
	 * horizontally.
	 */
	public final static int BOTH = 3;

	/**
	 * The anchoring constant that designates anchoring to the top of the display
	 * area
	 */
	public final static int TOP = 1;
	/**
	 * The anchoring constant that designates anchoring to the bottom of the display
	 * area
	 */
	public final static int BOTTOM = 2;
	private int vgap; // the vertical vgap between components...defaults to 5
	private int alignment; // LEFT, RIGHT, CENTER or BOTH...how the components are justified
	private int anchor; // TOP, BOTTOM or CENTER ...where are the components positioned in an overlarge
						// space

	// Constructors
	/**
	 * Constructs an instance of VerticalLayout with a vertical vgap of 5 pixels,
	 * left alignment and anchored to the top of the display area.
	 */
	public VerticalLayout()
	{
		this(5, LEFT, TOP);
	}

	/**
	 * Constructs a VerticalLayout instance with left alignment, anchored to
	 * the top with the specified vgap
	 *
	 * @param vgap An int value indicating the vertical seperation of the components
	 */
	public VerticalLayout(int vgap)
	{
		this(vgap, LEFT, TOP);
	}

	/**
	 * Constructs a VerticalLayout instance anchored to the top with the specified
	 * vgap and horizontal alignment
	 *
	 * @param vgap      An int value indicating the vertical seperation of the
	 *                  components
	 * @param alignment An int value which is one of
	 *                  <code>RIGHT, LEFT, CENTER, BOTH</code> for the horizontal
	 *                  alignment.
	 */
	public VerticalLayout(int vgap, int alignment)
	{
		this(vgap, alignment, TOP);
	}

	/**
	 * Constructs a VerticalLayout instance with the specified vgap, horizontal
	 * alignment and anchoring
	 *
	 * @param vgap      An int value indicating the vertical seperation of the
	 *                  components
	 * @param alignment An int value which is one of
	 *                  <code>RIGHT, LEFT, CENTER, BOTH</code> for the horizontal
	 *                  alignment.
	 * @param anchor    An int value which is one of
	 *                  <code>TOP, BOTTOM, CENTER</code> indicating where the
	 *                  components are to appear if the display area exceeds the
	 *                  minimum necessary.
	 */
	public VerticalLayout(int vgap, int alignment, int anchor)
	{
		this.vgap = vgap;
		this.alignment = alignment;
		this.anchor = anchor;
	}

	// ----------------------------------------------------------------------------
	private Dimension layoutSize(Container parent, boolean minimum)
	{
		Dimension dim = new Dimension(0, 0);
		Dimension d;
		synchronized (parent.getTreeLock())
		{
			int n = parent.getComponentCount();
			for (int i = 0; i < n; i++)
			{
				Component c = parent.getComponent(i);
				if (c.isVisible())
				{
					d = minimum ? c.getMinimumSize() : c.getPreferredSize();
					dim.width = Math.max(dim.width, d.width);
					dim.height += d.height;
					if (i > 0)
						dim.height += vgap;
				}
			}
		}
		Insets insets = parent.getInsets();
		dim.width += insets.left + insets.right;
		dim.height += insets.top + insets.bottom + Math.max(0, vgap + vgap);
		return dim;
	}

	// -----------------------------------------------------------------------------
	/**
	 * Lays out the container.
	 */
	@Override
	public void layoutContainer(Container parent)
	{
		Insets insets = parent.getInsets();
		synchronized (parent.getTreeLock())
		{
			int n = parent.getComponentCount();
			Dimension pd = parent.getSize();
			int y = insets.top;
			// work out the minimum height
			int lastNonZeroWeight = -1;
			for (int i = 0; i < n; i++)
			{
				Component c = parent.getComponent(i);
				Number weight = constraints.get(c);
				if (weight != null && weight.doubleValue() != 0)
					lastNonZeroWeight = i;
				Dimension d = c.getPreferredSize();
				y += d.height + vgap;
			}
			y -= vgap; // otherwise there's a vgap too many
			y += insets.bottom;
			int availableH = pd.height - y;
			int allocatedH = 0;
			// Work out the anchor paint
			if (anchor == TOP || lastNonZeroWeight != -1)
				y = insets.top;
			else if (anchor == CENTER)
				y = (pd.height - y) / 2;
			else
				y = pd.height - y - insets.bottom;
			// do layout
			for (int i = 0; i < n; i++)
			{
				Component c = parent.getComponent(i);
				Dimension d = c.getPreferredSize();
				int x = insets.left, wid = d.width, heig = d.height;

				Number weight = constraints.get(c);
				int ah;
				if (constraintSum == 0 || weight == null || weight.doubleValue() == 0)
					ah = 0;
				else if (i == lastNonZeroWeight)
					ah = availableH - allocatedH;
				else
					ah = (int) (availableH * weight.doubleValue() / constraintSum);
				heig += ah;
				allocatedH += ah;

				if (alignment == CENTER)
					x = (pd.width - d.width) / 2;
				else if (alignment == RIGHT)
					x = pd.width - d.width - insets.right;
				else if (alignment == BOTH)
					wid = pd.width - insets.left - insets.right;
				c.setBounds(x, y, wid, Math.max(0, heig));
				y += heig + vgap;
			}
		}
	}

	@Override
	public Dimension minimumLayoutSize(Container parent)
	{
		return layoutSize(parent, false);
	}

	@Override
	public Dimension preferredLayoutSize(Container parent)
	{
		return layoutSize(parent, false);
	}

	// ----------------------------------------------------------------------------
	/**
	 * Not used by this class
	 */
	@Override
	public void addLayoutComponent(String name, Component comp)
	{
	}

	// -----------------------------------------------------------------------------
	/**
	 * Not used by this class
	 */
	@Override
	public void removeLayoutComponent(Component comp)
	{
		if (constraints.containsKey(comp))
			constraintSum -= constraints.get(comp).doubleValue();
	}

	@Override
	public String toString()
	{
		return getClass().getName() + "[vgap=" + vgap + " align=" + alignment + " anchor=" + anchor + "]";
	}

	HashMap<Component, Number> constraints = new HashMap<Component, Number>();
	double constraintSum = 0;

	@Override
	public void addLayoutComponent(Component comp, Object constraints)
	{
		if (this.constraints.containsKey(comp))
			constraintSum -= this.constraints.remove(comp).doubleValue();
		if (constraints == null)
			return;
		if (!(constraints instanceof Number))
			throw new IllegalArgumentException(getClass().getSimpleName() + " constraint must be a number");
		this.constraints.put(comp, (Number) constraints);
		constraintSum += ((Number) constraints).doubleValue();
	}

	@Override
	public Dimension maximumLayoutSize(Container target)
	{
		return null;
	}

	@Override
	public float getLayoutAlignmentX(Container target)
	{
		return 0;
	}

	@Override
	public float getLayoutAlignmentY(Container target)
	{
		return 0;
	}

	@Override
	public void invalidateLayout(Container target)
	{
		// No-op
	}
}